# encoding: utf-8
from __future__ import unicode_literals
import abc
from itertools import groupby

from kinder.core.audit_log.models import MisLog
from kinder.core.dict.models import WorkType

MIS_LOG_INSERT_EVENT = 'insert'
MIS_LOG_FIELDS = ('object_id', 'object_value', 'isdata', 'timestamp')

WORK_TYPE_SHORT = WorkType.objects.get(code=WorkType.SHORT).name


class ModelLog(object):
    """
    Модель для работы с логами.

    Использование:

    ::

        # Определяем кастомную модель
        class CustomLog(ModelLog):
            model_name = 'Children'

        # Создаем рассчетный период
        period = Period('today', datetime.today(), datetime.today())

        # Получаем инстанс
        log = CustomLog({'id': 123}, period)

        # Возвращаем генератор значений за каждый день
        history_values = log.history_value([111, 222, 333])

        # Результат вычислений суммируется
        sum(history_values)  # 15
    """
    __metaclass__ = abc.ABCMeta

    @abc.abstractproperty
    def model_name(self):
        """Название модели по которой будет идти поиск изменений."""

    def __init__(self, unit, period):
        self.unit = unit
        self.period = period

    @staticmethod
    def records(period, model_name, object_ids):
        """Возвращает логи с нужными полями."""
        return MisLog.objects.filter(
            timestamp__range=(period.start, period.end),
            object_id__in=object_ids,
            object_type=model_name,
            event=MIS_LOG_INSERT_EVENT,
        ).values(*MIS_LOG_FIELDS)

    @staticmethod
    def events_on_date(events):
        """Возвращает события на рассчетную дату."""
        events_by_object = groupby(events, lambda rec: rec['object_id'])
        for object_id, object_events in events_by_object:
            # В день над объектом может быть несколько операций
            # берем последнюю в этот день операцию
            yield max(object_events, key=lambda rec: rec['timestamp'])

    def history_value(self, object_ids):
        """Возвращает количество событий на рассчетную дату."""
        events_by_date = groupby(
            self.records(self.period, self.model_name, object_ids),
            lambda rec: rec['timestamp'].date())
        for day, events in events_by_date:
            yield len(tuple(self.events_on_date(events)))


class DeclarationStatusLog(ModelLog):
    """Заявления в статусе Желает изменить ДОУ."""
    model_name = 'Declaration'


class DeclarationWorkTypeLog(ModelLog):
    """Количество заявлений с приоритетом кратковременное пребывание."""
    model_name = 'Declaration'


class DirectTemporaryLog(ModelLog):
    """Количество зачислений."""
    model_name = 'Direct'


class ParentsLog(ModelLog):
    """Количество родителей."""
    model_name = 'Delegate'


class ParentsSnilsLog(ModelLog):
    """Родители имеющие заполненный СНИЛС."""
    model_name = 'Delegate'
